/* (C) 2002 by Stealth
 */
#ifdef __KERNEL__
#define MODULE
#include <linux/stddef.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/sys.h>

#else
#include <stdio.h>
#include <string.h>
#endif

#include "ia32-decode.h"
#include "ia32-func-resolv.h"


static int is_return(ia32_instruction *instp)
{
	return (strstr(instp->opc.opcode->name, "ret") != NULL);
}


// relative calls can have displ_value == 0 but these
// is probably not a functioncall, so we dont return true then
static int is_relative_call(ia32_instruction *instp)
{
	return (strcmp(instp->opc.opcode->name, "call")  == 0 &&
	        instp->opc.displ_value != 0);
}


static int is_absolute_call(ia32_instruction *instp)
{
	return (strcmp(instp->opc.opcode->name, "call") == 0 &&
	        instp->opc.displ_size == 0);
}

static int holds_interesting_displacement(ia32_instruction *instp)
{
	return (strcmp(instp->opc.opcode->name, "mov") == 0 &&
		instp->opc.displ_size != 0 &&
		instp->opc.target_type == OP_TYPE_REG &&
		instp->opc.source_type != OP_TYPE_REG);
}


static int displacement2offset(unsigned int displ)
{
	int d = displ;	/* ugh! but must ... */
	return d + 5;
}


// Works only on real calls (no inlining etc ofcorse) and
// calls through global function pointers
void *ia32_function_resolve(void *start_func, int *walk_array, int windex)
{
	ia32_instruction *instp, inst;
	unsigned char *ptr = start_func;
	unsigned int i = 0, last_displacement = 0;

	if (!ptr)
		return NULL;

	for (;;) {
		instp = ia32_decode_instruction(&ptr[i], &inst);
		if (instp == NULL)
			break;

		// If theres is a displacement, save it for use
		// with absolute calls
		if (holds_interesting_displacement(instp)) {
			last_displacement = instp->opc.displ_value;
		} else if (!is_absolute_call(instp))
			last_displacement = 0;

	//	ia32_print(instp);
		if (!instp) {
#ifdef __KERNEL__
			printk("<1>invalid instruction at %d\n", i);
#else
			fprintf(stderr, "invalid instruction at %d\n", i);
			exit(1);
#endif
		}
		if (is_return(instp)) {
#ifdef __KERNEL__
			printk("<1>Invalid count at index %d\n", windex);
#else
			fprintf(stderr, "Invalid count at index %d\n", windex);
			exit(1);
#endif
		}
		if (is_relative_call(instp)) {
			walk_array[windex]--;

			// Walked down to function we have to follow
			if (walk_array[windex] == 0) {
				// If next element of walk_array is -1, we reached final
				// function even
				if (walk_array[windex+1] == -1)
					return &ptr[i]+displacement2offset(instp->opc.displ_value);
				// if not we need to go deeper
				else {
					int offset = displacement2offset(instp->opc.displ_value);
					return ia32_function_resolve(&ptr[i]+offset, walk_array, windex+1);
				}
			}
		// only follow absolute call when we have valid displacement
		} else if (is_absolute_call(instp) && last_displacement) {
			walk_array[windex]--;
			if (walk_array[windex+1] == -1)
				return (void*)*((int*)last_displacement);
			else {
				return ia32_function_resolve((void*)*(int*)last_displacement, walk_array, windex+1);
			}
		}
		i += instp->length;
	}
	return NULL;
}


void *ia32_find_absolute_call(void *start_func)
{
	ia32_instruction *instp, inst;
	unsigned int i = 0, last_displacement = 0;
	unsigned char *ptr = start_func;

	if (!ptr)
		return NULL;

	for (;;) {
		instp = ia32_decode_instruction(&ptr[i], &inst);
		if (instp == NULL)
			break;

		if (holds_interesting_displacement(instp)) {
			last_displacement = instp->opc.displ_value;
		} else if (!is_absolute_call(instp))
			last_displacement = 0;

		if (is_return(instp))
			break;
		else if (is_absolute_call(instp) && last_displacement)
			return (void*)*((int*)last_displacement);
			
		i += instp->length;
	}
	return NULL;
	
}


// only relative calls count
void ia32_call_pattern(void *start_func, int *arr, size_t asize, int depth, int firstcall)
{
	static int aindex = 0;
	void *funcarr[1024];
	ia32_instruction *instp, inst;
	unsigned int ncalls = 0, i = 0;
	unsigned char *ptr = start_func;

#ifdef __KERNEL__
	if (((unsigned int)ptr & 0xf0000000) < 0xc0000000) {
		printk("!!! %p\n", ptr); return;
	}
#endif
	if (firstcall)
		aindex = 0;

	if (depth < 0 || aindex >= asize)
		return;

	for (ncalls = 0; ncalls < 1024;) {
#ifdef __KERNEL__
		if (((unsigned int)&ptr[i] & 0xf0000000) < 0xc0000000) {
			break;
		}
#endif


		instp = ia32_decode_instruction(&ptr[i], &inst);

		if (instp == NULL)
			return;
		//printk("%p: ", &ptr[i]);
		//ia32_print(instp);
	
		if (is_return(instp))
			break;
		else if (is_relative_call(instp)) {
			int offset = displacement2offset(instp->opc.displ_value);
//		printk("-> %d %p: %d %p\n", ncalls++, &ptr[i], offset, &ptr[i]+offset);
			funcarr[ncalls++] = &ptr[i]+offset;
		}
		i += instp->length;
	}
	arr[aindex++] = ncalls;
	for (i = 0; i < ncalls; ++i) {
		ia32_call_pattern(funcarr[i], arr, asize, depth-1, 0);
	}
	
	return;
}

